#
# Copyright (C) 2021 Luis López <luis@cuarentaydos.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
# USA.


import functools
import hashlib
import hmac
import logging
import time
import xml.dom.minidom

import requests
import xmltodict

from .const import (
    DEFAULT_MODULE_ID,
    DEFAULT_PORT,
    DEFAULT_REQUEST_TIMEOUT,
    DEFAULT_SESSION_LIFETIME,
    DEFAULT_USERNAME,
)
from .helpers import auth_required

_LOGGER = logging.getLogger(__name__)


def hex_hmac_md5(a: str, b: str) -> str:
    return hmac.new(a.encode("ascii"), b.encode("ascii"), hashlib.md5).hexdigest()


class SoapClient:
    HNAP1_XMLNS = "http://purenetworks.com/HNAP1/"
    HNAP_METHOD = "POST"
    HNAP_BODY_ENCODING = "UTF8"
    HNAP_LOGIN_METHOD = "Login"
    HNAP_AUTH = {
        "challenge": "",
        "cookie": "",
        "private_key": "",
        "public_key": "",
        "password": None,
        "result": "",
        "url": "http://{hostname}:{port}/HNAP1/",
        "username": None,
    }

    def __init__(
        self,
        hostname,
        password,
        username=DEFAULT_USERNAME,
        port=DEFAULT_PORT,
        request_timeout=DEFAULT_REQUEST_TIMEOUT,
        session_lifetime=DEFAULT_SESSION_LIFETIME,
    ):
        self._hostname = hostname
        self._port = port
        self._request_timeout = request_timeout
        self._session_lifetime = session_lifetime

        self.HNAP_AUTH = self.HNAP_AUTH.copy()
        self.HNAP_AUTH["url"] = self.HNAP_AUTH["url"].format(
            hostname=hostname, port=port
        )
        self.HNAP_AUTH["username"] = username
        self.HNAP_AUTH["password"] = password
        self._authenticated = 0

    @property
    def hostname(self):
        return self._hostname

    @property
    def port(self):
        return self._port

    @property
    def username(self):
        return self.HNAP_AUTH["username"]

    @property
    def password(self):
        return self.HNAP_AUTH["password"]

    def _build_method_envelope(self, method, **parameters):
        parameters_xml = "\n".join(
            [f"     <{k}>{v}</{k}>" for (k, v) in parameters.items()]
        )
        return (
            '<?xml version="1.0" encoding="utf-8"?>'
            "<soap:Envelope "
            '  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            '  xmlns:xsd="http://www.w3.org/2001/XMLSchema" '
            '  xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">'
            "  <soap:Body>"
            f'   <{method} xmlns="{self.HNAP1_XMLNS}">'
            f"     {parameters_xml}"
            f"   </{method}>"
            "  </soap:Body>"
            "</soap:Envelope>"
        )

    def _save_login_result(self, body):
        doc = xml.dom.minidom.parseString(body)
        self.HNAP_AUTH["result"] = doc.getElementsByTagName(
            f"{self.HNAP_LOGIN_METHOD}Result"
        )[0].firstChild.nodeValue

        for tag, key in [
            ("Challenge", "challenge"),
            ("PublicKey", "public_key"),
            ("Cookie", "cookie"),
        ]:
            elements = doc.getElementsByTagName(tag)
            self.HNAP_AUTH[key] = elements[0].firstChild.nodeValue

        self.HNAP_AUTH["private_key"] = hex_hmac_md5(
            self.HNAP_AUTH["public_key"] + self.HNAP_AUTH["password"],
            self.HNAP_AUTH["challenge"],
        ).upper()

    def _getHNAP_auth(self, soap_action, private_key):
        time_stamp = int(time.mktime(time.localtime()))
        auth = hex_hmac_md5(private_key, str(time_stamp) + soap_action)
        ret = auth.upper() + " " + str(time_stamp)

        return ret

    def call_raw(self, method, **parameters):
        req_url = self.HNAP_AUTH["url"]
        req_method = self.HNAP_METHOD
        req_headers = {
            "Content-Type": "text/xml; charset=utf-8",
            "SOAPAction": f'"{self.HNAP1_XMLNS}{method}"',
            "HNAP_AUTH": self._getHNAP_auth(
                f'"{self.HNAP1_XMLNS}{method}"',
                self.HNAP_AUTH["private_key"],
            ),
            "Cookie": "uid=" + self.HNAP_AUTH["cookie"],
        }

        resp = requests.request(
            method=req_method,
            url=req_url,
            headers=req_headers,
            data=self._build_method_envelope(method, **parameters),
            timeout=self._request_timeout,
        )

        if resp.status_code != 200:
            raise MethodCallError(
                f"Invalid status code: {resp.status_code}", resp.status_code
            )
        return resp.text

    def call(self, method, **parameters):
        parsed = xmltodict.parse(self.call_raw(method, **parameters))
        try:
            res = parsed["soap:Envelope"]["soap:Body"][f"{method}Response"][
                f"{method}Result"
            ]
            if res.lower() not in ("ok", "success"):
                raise MethodCallError(f"{method} returned {res}")

        except KeyError:
            raise MethodCallError(f"Missing {method}Result key")

        return parsed["soap:Envelope"]["soap:Body"][f"{method}Response"]

    def authenticate(self, force=False):
        if self.is_authenticated() and not force:
            _LOGGER.debug("Client already authenticated")
            return

        url = self.HNAP_AUTH["url"]
        method = self.HNAP_METHOD
        data = self._build_method_envelope(
            self.HNAP_LOGIN_METHOD,
            Action="request",
            Username=self.HNAP_AUTH["username"],
            LoginPassword="",
            Captcha="",
        )
        headers = {
            "Content-Type": "text/xml; charset=utf-8",
            "SOAPAction": '"' + self.HNAP1_XMLNS + self.HNAP_LOGIN_METHOD + '"',
        }

        resp = requests.request(
            method=method,
            url=url,
            data=data,
            headers=headers,
            timeout=self._request_timeout,
        )

        if resp.status_code != 200:
            raise AuthenticationError(
                f"Invalid response while login: {resp.status_code} ({resp.text})"
            )

        self._save_login_result(resp.text)

        # Phase 2
        login_password = hex_hmac_md5(
            self.HNAP_AUTH["private_key"], self.HNAP_AUTH["challenge"]
        ).upper()

        res = self.call(
            self.HNAP_LOGIN_METHOD,
            Action="login",
            Username=self.HNAP_AUTH["username"],
            LoginPassword=login_password,
            Captcha="",
        )

        if res["LoginResult"] != "success":
            raise AuthenticationError(res["LoginResult"])

        self._authenticated = time.monotonic()

    def is_authenticated(self):
        return (self._authenticated > 0) and (
            (time.monotonic() - self._authenticated) <= self._session_lifetime
        )

    def _inspect_device(self, **kwargs):
        def _unwrap_string_ordered_dict(data):
            ret = []

            for modtype, modname in data.items():
                if modtype != "string":
                    _LOGGER.warning(f"Unknow type class: ({modtype}, {modname}")
                    continue

                if isinstance(modname, str):
                    ret.append(modname)

                elif isinstance(modname, list):
                    ret.extend(modname)
                else:
                    _LOGGER.error(f"Unknow value instance: ({modname!r}")

            return ret

        info = dict(self.call("GetDeviceSettings", **kwargs))

        # Rewrite some keys

        # Rewrite some keys
        info["ModuleTypes"] = (
            _unwrap_string_ordered_dict(info["ModuleTypes"])
            if "ModuleTypes" in info
            else []
        )
        info["SOAPActions"] = _unwrap_string_ordered_dict(info["SOAPActions"])

        return info

    @auth_required
    def device_info(self, **kwargs):
        info = self._inspect_device(**kwargs)

        for k in ["@xmlns", "SOAPActions", "GetDeviceSettingsResult"]:
            info.pop(k, None)

        return info

    @auth_required
    def device_actions(self, **kwargs):
        idx = len(self.HNAP1_XMLNS)

        info = self._inspect_device(**kwargs)
        actions = info["SOAPActions"]
        actions = (x for x in actions if x.startswith(self.HNAP1_XMLNS))
        actions = (x[idx:] for x in actions)

        return list(actions)

    @auth_required
    def module_actions(self, *, ModuleID=DEFAULT_MODULE_ID, **kwargs):
        resp = self.call("GetModuleSOAPActions", ModuleID=ModuleID, **kwargs)
        actions = resp["ModuleSOAPList"]["SOAPActions"]["Action"]

        return actions


class ClientError(Exception):
    pass


class AuthenticationError(ClientError):
    pass


class MethodCallError(ClientError):
    pass



#
# Copyright (C) 2021 Luis López <luis@cuarentaydos.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
# USA.


import logging
from datetime import datetime
from enum import Enum

from .const import DEFAULT_MODULE_ID, DEFAULT_PORT, DEFAULT_USERNAME
from .helpers import auth_required
from .soapclient import MethodCallError, SoapClient

_LOGGER = logging.getLogger(__name__)


def DeviceFactory(
    *,
    client=None,
    hostname=None,
    password=None,
    username=DEFAULT_USERNAME,
    port=DEFAULT_PORT,
):
    client = client or SoapClient(
        hostname=hostname, password=password, username=username, port=port
    )
    info = client.device_info()

    module_types = info["ModuleTypes"]
    if not isinstance(module_types, list):
        module_types = [module_types]

    if "Audio Renderer" in module_types:
        cls = Siren

    elif "Camera" in module_types:
        # Other posible values for camera (needs testing):
        # 'Optical Recognition', 'Environmental Sensor', 'Camera'
        cls = Camera

    elif "Motion Sensor" in module_types:
        cls = Motion

    else:
        raise TypeError(module_types)

    return cls(client=client)


class Device:
    MODULE_TYPE: str

    def __init__(
        self,
        *,
        client=None,
        hostname=None,
        password=None,
        username=DEFAULT_USERNAME,
        port=DEFAULT_PORT,
        module_id=DEFAULT_MODULE_ID,
    ):
        self.client = client or SoapClient(
            hostname=hostname, password=password, username=username, port=port
        )
        self.module_id = module_id

        self._info = None

    @property
    def info(self):
        if not self._info:
            self._info = self.client.device_info()

        return self._info

    def call(self, *args, **kwargs):
        kwargs["ModuleID"] = self.module_id
        return self.client.call(*args, **kwargs)

    def is_authenticated(self):
        return self.client.is_authenticated()

    def authenticate(self):
        return self.client.authenticate()


class Camera(Device):
    MODULE_TYPE = "Camera"
    DEFAULT_SCHEMA = "http://"
    DEFAULT_STREAM_PATH = "/play1.sdp"
    DEFAULT_PICTURE_PATH = "/image/jpeg.cgi"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._base_url = (
            {self.DEFAULT_SCHEMA}
            + f"{self.client.username.lower()}:{self.client.password}@"
            + f"{self.client.hostname}:{self.client.port}"
        )

    @property
    def stream_url(self):
        return f"{self._base_url}{self.DEFAULT_STREAM_PATH}"

    @property
    def picture_url(self):
        return f"{self._base_url}{self.DEFAULT_PICTURE_PATH}"


class Motion(Device):
    MODULE_TYPE = "Motion Sensor"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._backoff = None

    @property
    def backoff(self):
        if self._backoff is None:
            resp = self.call("GetMotionDetectorSettings")
            try:
                self._backoff = int(resp["Backoff"])
            except (KeyError, ValueError, TypeError):
                # Return the default value for tested devices but don't store
                # to force retry
                return 30

        return self._backoff

    # @backoff.setter
    # def backoff(self, seconds):
    #     self.call(
    #         "SetMotionDetectorSettings", Backoff=self._backoff
    #     )
    #     _LOGGER.warning("set backoff property has no effect")

    # def authenticate(self):
    #     super().authenticate()

    #     res = self.call("GetMotionDetectorSettings")
    #     try:
    #         self._backoff = int(res["Backoff"])
    #     except (ValueError, TypeError, KeyError):
    #         _LOGGER.warning("Unable to get delta from device")

    @auth_required
    def get_latest_detection(self):
        res = self.call("GetLatestDetection")
        return datetime.fromtimestamp(float(res["LatestDetectTime"]))

    @auth_required
    def is_active(self):
        now = datetime.now()
        diff = (now - self.get_latest_detection()).total_seconds()

        return diff <= self.backoff


class Router(Device):
    # NOT tested
    # See https://github.com/waffelheld/dlink-device-tracker/blob/master/custom_components/dlink_device_tracker/dlink_hnap.py#L95  # noqa: E501
    MODULE_TYPE = "check-module-types-for-router"

    @auth_required
    def get_clients(self):
        res = self.call("GetClientInfo")
        clients = res["ClientInfoLists"]["ClientInfo"]

        # Filter out offline clients
        # clients = [x for x in clients if x["Type"] != "OFFLINE"]

        ret = [
            {
                "name": client["DeviceName"],
                "nickName": client["NickName"],
                "is_connected": client["Type"] == "OFFLINE" and 0 or 1,
                "mac": client["MacAddress"],
            }
            for client in clients
        ]
        return ret


class SirenSound(Enum):
    EMERGENCY = 1
    FIRE = 2
    AMBULANCE = 3
    POLICE = 4
    DOOR_CHIME = 5
    BEEP = 6

    @classmethod
    def fromstring(cls, s):
        s = s.upper()
        for c in ["-", " ", "."]:
            s = s.replace(c, "_")

        return getattr(cls, s)


class Siren(Device):
    MODULE_TYPE = "Audio Renderer"

    @auth_required
    def is_playing(self):
        res = self.call("GetSirenAlarmSettings")
        return res["IsSounding"] == "true"

    @auth_required
    def play(self, sound=SirenSound.EMERGENCY, volume=100, duration=60):
        ret = self.call(
            "SetSoundPlay",
            SoundType=sound.value,
            Volume=volume,
            Duration=duration,
        )
        if ret["SetSoundPlayResult"] != "OK":
            raise MethodCallError(f"Unable to play. Response: {ret}")

    @auth_required
    def beep(self, volume=100, duration=1):
        return self.play(sound=SirenSound.BEEP, duration=duration, volume=volume)

    @auth_required
    def stop(self):
        ret = self.call("SetAlarmDismissed")

        if ret["SetAlarmDismissedResult"] != "OK":
            raise MethodCallError(f"Unable to stop. Response: {ret}")


class Water(Device):
    # NOT tested
    MODULE_TYPE = "check-module-types-for-water-detector"

    @auth_required
    def is_active(self):
        ret = self.call("GetWaterDetectorState")
        return ret.get("IsWater") == "true"



#
# Copyright (C) 2021 Luis López <luis@cuarentaydos.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
# USA.


import argparse
import logging
import os
import pprint
import sys
import xml.dom.minidom

import requests

from .soapclient import SoapClient

OUTPUT_TMPL = """
Device info
===========
{info}

Device actions
==============
{device_actions}

Module actions
==============
{module_actions}
"""


def main():
    logging.basicConfig()
    logging.getLogger("hnap").setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--hostname",
        required=True,
        metavar="hostname",
    )
    parser.add_argument(
        "--password",
        required=True,
        metavar="password",
        default=os.environ.get("HNAP_PASSWORD", ""),
    )
    parser.add_argument(
        "--username",
        default="admin",
        metavar="username",
    )
    parser.add_argument(
        "--call",
        nargs=1,
        default=[],
        metavar="MethodName",
        help=(
            "If you are calling a SOAP action you must pass, at least, "
            "`--param Module 1` and, maybe, `--param Controller 1`."
        ),
    )
    parser.add_argument(
        "--param",
        action="append",
        nargs=2,
        default=[],
        dest="params",
        metavar=("ParamName", "Value"),
        help="Params to pass to call. Multiple params can be passed.",
    )
    args = parser.parse_args()

    client = SoapClient(
        hostname=args.hostname,
        username=args.username,
        password=args.password,
    )

    try:
        client.authenticate()

    except requests.ReadTimeout:
        print(
            f"{args.hostname}: read timeout error "
            "(Is device stuck? try rebooting it)",
            file=sys.stderr,
        )
        sys.exit(1)
    except requests.ConnectionError:
        print(
            f"{args.hostname}: connection error (offline device? wrong hostname?)",
            file=sys.stderr,
        )
        sys.exit(1)

    if args.call:
        params = dict(args.params)
        resp = client.call_raw(args.call[0], **params)
        if not resp:
            print("Empty (or unsupported) response")
            return

        dom = xml.dom.minidom.parseString(resp)
        print(dom.toprettyxml())

    else:
        print(
            OUTPUT_TMPL.format(
                info=pprint.pformat(client.device_info()),
                device_actions=pprint.pformat(client.device_actions()),
                module_actions=pprint.pformat(client.module_actions()),
            ).strip()
        )


if __name__ == "__main__":
    main()
